#!/usr/bin/env python
# encoding: utf-8
"""
tophat.py

Created by Cole Trapnell on 2008-12-25.
Copyright (c) 2008 Cole Trapnell. All rights reserved.
"""

import sys
try:
    import psyco
    psyco.full()
except ImportError:
    pass

import sys
import getopt
import subprocess
import errno
import os
import tempfile
import warnings
import shutil
import copy
from datetime import datetime, date, time

# use_message = '''
# TopHat maps short sequences from spliced transcripts to whole genomes.
# 
# Usage:
#     tophat [options] <bowtie_index> <reads1[,reads2,...,readsN]> [reads1[,reads2,...,readsN]]
#     
# Options:
#     -o/--output-dir                <string>    [ default: ./tophat_out ]
#     -a/--min-anchor                <int>       [ default: 8     ]
#     -m/--splice-mismatches         <0-2>       [ default: 0     ]
#     -i/--min-intron                <int>       [ default: 50    ]
#     -I/--max-intron                <int>       [ default: 20000 ]
#     -g/--max-gene-family           <int>       [ default: 40    ]
#     -F/--min-isoform-fraction      <float>     [ default: 0.15  ]
#     -X/--solexa-quals                          
#     -p/--num-threads               <int>       [ default: 1     ]
#     -G/--GFF                       <filename>
#     -j/--raw-juncs                 <filename>
#     -r/--mate-inner-dist           <int>       [ default: 75    ]
#     --mate-std-dev                 <int>       [ default: 20    ]
#     --no-novel-juncs                           
#     --no-gff-juncs                             
#     --no-coverage-search                       
#     --no-closure-search
#     --fill-gaps                        
# '''

use_message = '''
TopHat maps short sequences from spliced transcripts to whole genomes.

Usage:
    tophat [options] <bowtie_index> <reads1[,reads2,...,readsN]> [reads1[,reads2,...,readsN]]
    
Options:
    -o/--output-dir                <string>    [ default: ./tophat_out ]
    -a/--min-anchor                <int>       [ default: 8            ]
    -m/--splice-mismatches         <0-2>       [ default: 0            ]
    -i/--min-intron                <int>       [ default: 50           ]
    -I/--max-intron                <int>       [ default: 500000       ]
    -g/--max-multihits             <int>       [ default: 40           ]
    -F/--min-isoform-fraction      <float>     [ default: 0.15         ]
    --solexa-quals                          
    --solexa1.3-quals
    -p/--num-threads               <int>       [ default: 1            ]
    -G/--GFF                       <filename>
    -j/--raw-juncs                 <filename>
    -r/--mate-inner-dist           <int>       
    --mate-std-dev                 <int>       [ default: 20           ]
    --no-novel-juncs                           
    --no-gff-juncs                             
    --no-coverage-search
    --coverage-search                                              
    --no-closure-search
    --closure-search
    --fill-gaps        
    --microexon-search
    --butterfly-search
    --no-butterfly-search
    --keep-tmp
    
Advanced Options:

    --segment-mismatches           <int>       [ default: 2            ]
    --segment-length               <int>       [ default: 25           ]
    --min-closure-exon             <int>       [ default: 100          ]
    --min-closure-intron           <int>       [ default: 50           ]
    --max-closure-intron           <int>       [ default: 5000         ]
    --min-coverage-intron          <int>       [ default: 50           ]
    --max-coverage-intron          <int>       [ default: 20000        ]
    --min-segment-intron           <int>       [ default: 50           ]
    --max-segment-intron           <int>       [ default: 500000       ]    
'''

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

output_dir = "./tophat_out/"
logging_dir = output_dir + "logs/"
run_log = None

tmp_dir = output_dir + "tmp/"
bin_dir = sys.path[0] + "/"

#ok_str = "\t\t\t\t[OK]\n"
fail_str = "\t[FAILED]\n"


class TopHatParams:
    class SpliceConstraints:
        def __init__(self, 
                     min_anchor_length,
                     min_intron_length,
                     max_intron_length, 
                     splice_mismatches,
                     min_isoform_fraction):
            self.min_anchor_length = min_anchor_length
            self.min_intron_length = min_intron_length
            self.max_intron_length = max_intron_length
            self.splice_mismatches = splice_mismatches
            self.min_isoform_fraction = min_isoform_fraction
        
        def parse_options(self, opts):
            for option, value in opts:
                if option in ("-m", "--splice-mismatches"):
                    self.splice_mismatches = int(value)
                if option in ("-a", "--min-anchor"):
                    self.min_anchor_length = int(value)
                if option in ("-F", "--min-isoform-fraction"):
                    self.min_isoform_fraction = float(value)
                if option in ("-i", "--min-intron-length"):
                    self.min_intron_length = int(value)
                if option in ("-I", "--max-intron-length"):
                    self.max_intron_length = int(value)
        
        def check(self):
            if self.splice_mismatches not in [0,1,2]:
                print >> sys.stderr, "Error: arg to --splice-mismatches must be 0, 1, or 2"
                exit(1)
            if self.min_anchor_length < 4:
                print >> sys.stderr, "Error: arg to --min-anchor-len must be greater than 4"
                exit(1)
            if self.min_isoform_fraction < 0.0 or self.min_isoform_fraction > 1.0:
                print >> sys.stderr, "Error: arg to --min-isoform-fraction must be between 0.0 and 1.0"
                exit(1)
            if self.min_intron_length <= 0:
                print >> sys.stderr, "Error: arg to --min-intron-length must be greater than 0"
                exit(1)                    
            if self.max_intron_length <= 0:
                print >> sys.stderr, "Error: arg to --max-intron-length must be greater than 0"
                exit(1)
        
    class SystemParams:
        def __init__(self,
                     bowtie_threads,
                     keep_tmp):
            self.bowtie_threads = bowtie_threads
            self.keep_tmp = keep_tmp
            
        def parse_options(self, opts):
            for option, value in opts:
                if option in ("-p", "--num-threads"):
                    self.bowtie_threads = int(value)
                if option in ("--keep-tmp"):
                    self.keep_tmp = True
        
        def check(self):
            pass
        
    
    class ReadParams:
        def __init__(self,
                     quality_scale,
                     seed_length,
                     reads_format,
                     mate_inner_dist,
                     mate_inner_dist_std_dev):
            self.quality_scale = quality_scale
            self.seed_length = seed_length
            self.reads_format = reads_format
            self.mate_inner_dist = mate_inner_dist
            self.mate_inner_dist_std_dev = mate_inner_dist_std_dev
            
        def parse_options(self, opts):
            for option, value in opts:
                if option in ("--solexa-quals", "--solexa1.3-quals"):
                    self.quality_scale = option
                if option in ("-s", "--seed-length"):
                    self.seed_length = int(value)
                if option in ("-r", "--mate-inner-dist"):
                    self.mate_inner_dist = int(value)
                if option == "--mate-std-dev":
                    self.mate_inner_dist_std_dev = int(value)
                    
        def check(self):
            if self.seed_length != None and self.seed_length < 20:
                print >> sys.stderr, "Error: arg to --seed-length must be at least 20"
                exit(1)
            if self.mate_inner_dist != None and self.mate_inner_dist < 0:
                print >> sys.stderr, "Error: arg to --mate-inner-dist must at least 0"
                exit(1)
            if self.mate_inner_dist_std_dev != None and self.mate_inner_dist_std_dev < 0:
                print >> sys.stderr, "Error: arg to --mate-std-dev must at least 0"
                exit(1)
                
    class SearchParams:
        def __init__(self,
                     min_closure_exon,
                     min_closure_intron,
                     max_closure_intron,
                     min_coverage_intron,
                     max_coverage_intron,
                     min_segment_intron,
                     max_segment_intron):
                     
             self.min_closure_exon_length = min_closure_exon
             self.min_closure_intron_length = min_closure_intron
             self.max_closure_intron_length = max_closure_intron
             self.min_coverage_intron_length = min_coverage_intron
             self.max_coverage_intron_length = max_coverage_intron
             self.min_segment_intron_length = min_segment_intron
             self.max_segment_intron_length = max_segment_intron

        def parse_options(self, opts):
            for option, value in opts:
                if option == "--min-closure-exon":
                    self.min_closure_exon_length = int(value)
                if option == "--min-closure-intron":
                    self.min_closure_intron_length = int(value)
                if option == "--max-closure-intron":
                    self.max_closure_intron_length = int(value)
                if option == "--min-coverage-intron":
                    self.min_coverage_intron_length = int(value)
                if option == "--max-coverage-intron":
                    self.max_coverage_intron_length = int(value)
                if option == "--min-segment-intron":
                    self.min_segment_intron_length = int(value)
                if option == "--max-segment-intron":
                    self.max_segment_intron_length = int(value)

        def check(self):
            if self.min_closure_exon_length < 0:
                print >> sys.stderr, "Error: arg to --min-closure-exon must be at least 20"
                exit(1)
            if self.min_closure_intron_length < 0:
                print >> sys.stderr, "Error: arg to --min-closure-intron must be at least 20"
                exit(1)
            if self.max_closure_intron_length < 0:
                print >> sys.stderr, "Error: arg to --max-closure-intron must be at least 20"
                exit(1)
            if self.min_coverage_intron_length < 0:
                print >> sys.stderr, "Error: arg to --min-coverage-intron must be at least 20"
                exit(1)
            if self.max_coverage_intron_length < 0:
                print >> sys.stderr, "Error: arg to --max-coverage-intron must be at least 20"
                exit(1)
            if self.min_segment_intron_length < 0:
                print >> sys.stderr, "Error: arg to --min-segment-intron must be at least 20"
                exit(1)
            if self.max_segment_intron_length < 0:
                print >> sys.stderr, "Error: arg to --max-segment-intron must be at least 20"
                exit(1)
                    
    def __init__(self):        
        self.splice_constraints = self.SpliceConstraints(8,     # min_anchor 
                                                         50,    # min_intron
                                                         500000, # max_intron
                                                         0,     # splice_mismatches
                                                         0.15)  # min_isoform_frac
        
        self.read_params = self.ReadParams("--phred33-quals",               # solexa_scale
                                           None,                # seed_length
                                           "fastq",             # quality_format
                                           None,                # mate inner distance
                                           20)                  # mate inner dist std dev
        
        self.system_params = self.SystemParams(1,               # bowtie_threads
                                               False)           # keep_tmp   
                                               
        self.search_params = self.SearchParams(100,             # min_closure_exon_length
                                               50,              # min_closure_intron_length
                                               5000,            # max_closure_intron_length
                                               50,              # min_coverage_intron_length
                                               20000,           # max_coverage_intron_length
                                               50,              # min_segment_intron_length
                                               500000)          # max_segment_intron_length                          
        
        self.gff_annotation = None
        self.raw_junctions = None
        self.find_novel_juncs = True
        self.find_GFF_juncs = True
        self.skip_check_reads = False
        self.max_hits = 40
        self.segment_length = 25
        self.segment_mismatches = 2
        self.closure_search = None
        self.coverage_search = None
        self.microexon_search = False
        self.butterfly_search = None
        
    def check(self):
        self.splice_constraints.check()
        self.read_params.check()
        self.system_params.check()
       
        if self.segment_length <= 4:
            print >> sys.stderr, "Error: arg to --segment-length must at least 4"
            exit(1)
        if self.segment_mismatches < 0 or self.segment_mismatches > 3:
            print >> sys.stderr, "Error: arg to --segment-mismatches must in [0, 3]"
            exit(1)
        
        self.search_params.max_closure_intron_length = min(self.splice_constraints.max_intron_length,
                                                           self.search_params.max_closure_intron_length)
        
        self.search_params.max_segment_intron_length = min(self.splice_constraints.max_intron_length,
                                                           self.search_params.max_segment_intron_length)

        self.search_params.max_coverage_intron_length = min(self.splice_constraints.max_intron_length,
                                                            self.search_params.max_coverage_intron_length)
        
    def cmd(self):
        cmd = ["--fastq",
               "--min-anchor", str(self.splice_constraints.min_anchor_length),
               "--splice-mismatches", str(self.splice_constraints.splice_mismatches),
               "--min-report-intron", str(self.splice_constraints.min_intron_length),
               "--max-report-intron", str(self.splice_constraints.max_intron_length),
               "--min-isoform-fraction", str(self.splice_constraints.min_isoform_fraction),
               "--output-dir", output_dir,
               "--max-multihits", str(self.max_hits),
               "--segment-length", str(self.segment_length),
               "--segment-mismatches", str(self.segment_mismatches),
               "--min-closure-exon", str(self.search_params.min_closure_exon_length),
               "--min-closure-intron", str(self.search_params.min_closure_intron_length),
               "--max-closure-intron", str(self.search_params.max_closure_intron_length),
               "--min-coverage-intron", str(self.search_params.min_coverage_intron_length),
               "--max-coverage-intron", str(self.search_params.max_coverage_intron_length),
               "--min-segment-intron", str(self.search_params.min_segment_intron_length),
               "--max-segment-intron", str(self.search_params.max_segment_intron_length)]
               
        if self.read_params.mate_inner_dist != None:
            cmd.extend(["--inner-dist-mean", str(self.read_params.mate_inner_dist),
                        "--inner-dist-std-dev", str(self.read_params.mate_inner_dist_std_dev)])
        if self.gff_annotation != None:
            cmd.extend(["--gff-annotations", str(self.gff_annotation)])
        if self.closure_search == False:
            cmd.append("--no-closure-search")
        if self.coverage_search == False:
            cmd.append("--no-coverage-search")
        if self.microexon_search == False:
            cmd.append("--no-microexon-search")            
        if self.butterfly_search == True:
            cmd.append("--butterfly-search")
        return cmd
        
    def parse_options(self, argv):
        try:
            opts, args = getopt.getopt(argv[1:], "hvp:m:F:a:i:I:G:r:o:j:g:", 
                                        ["version",
                                         "help",  
                                         "output-dir=",
                                         "solexa-quals",
                                         "solexa1.3-quals",
                                         "num-threads=",
                                         "splice-mismatches=",
                                         "max-multihits=",
                                         "min-isoform-fraction=",
                                         "min-anchor-length=",
                                         "min-intron-length=",
                                         "max-intron-length=",
                                         "GFF=",
                                         "raw-juncs=",
                                         "no-novel-juncs",
                                         "no-gff-juncs",
                                         "skip-check-reads",
                                         "mate-inner-dist=",
                                         "mate-std-dev=",
                                         "no-closure-search",
                                         "no-coverage-search",
                                         "closure-search",
                                         "coverage-search",
                                         "microexon-search",
                                         "min-closure-exon=",
                                         "min-closure-intron=",
                                         "max-closure-intron=",
                                         "min-coverage-intron=",
                                         "max-coverage-intron=",
                                         "min-segment-intron=",
                                         "max-segment-intron=",
                                         "segment-length=",
                                         "segment-mismatches=",
                                         "butterfly-search",
                                         "no-butterfly-search",
                                         "keep-tmp"])
        except getopt.error, msg:
            raise Usage(msg)
        
        self.splice_constraints.parse_options(opts)
        self.system_params.parse_options(opts)
        self.read_params.parse_options(opts)
        self.search_params.parse_options(opts)
        
        # option processing
        for option, value in opts:
            if option in ("-v", "--version"):
                print "TopHat v%s" % (get_version())
                exit(0)
            if option in ("-h", "--help"):
                raise Usage(use_message)
            if option in ("-g", "--max-gene-family"):
                self.max_hits = int(value)
            if option in ("-G", "--GFF"):
                self.gff_annotation = value
            if option in ("-j", "--raw-juncs"):
                self.raw_junctions = value
            if option == "--no-novel-juncs":
                self.find_novel_juncs = False
            if option == "--no-gff-juncs":
                self.find_GFF_juncs = False
            if option == "--skip-check-reads":
                self.skip_check_reads = True
            if option == "--no-coverage-search":
                self.coverage_search = False
            if option == "--no-closure-search":
                self.closure_search = False
            if option == "--coverage-search":
                self.coverage_search = True
            if option == "--closure-search":
                self.closure_search = True
            if option == "--microexon-search":
                self.microexon_search = True
            if option == "--butterfly-search":
                self.butterfly_search = True
            if option == "--no-butterfly-search":
                self.butterfly_search = False
            if option == "--segment-length":
                self.segment_length = int(value)
            if option == "--segment-mismatches":
                self.segment_mismatches = int(value)
            if option in ("-o", "--output-dir"):
                global output_dir
                global logging_dir
                global tmp_dir
                output_dir = value + "/"
                logging_dir = output_dir + "logs/"
                tmp_dir = output_dir + "tmp/"
            
        if len(args) < 2:
            raise Usage(use_message)
        return args

def right_now():
    curr_time = datetime.now()
    return curr_time.strftime("%c")

def prepare_output_dir():
    
    print >> sys.stderr, "[%s] Preparing output location %s" % (right_now(), output_dir)
    if os.path.exists(output_dir):
        pass
    else:        
        os.mkdir(output_dir)
        
    if os.path.exists(logging_dir):
        pass
    else:        
        os.mkdir(logging_dir)
        
    if os.path.exists(tmp_dir):
        pass
    else:        
        os.mkdir(tmp_dir)

def check_bowtie_index(idx_prefix):
    print >> sys.stderr, "[%s] Checking for Bowtie index files" % right_now()
    
    idx_fwd_1 = idx_prefix + ".1.ebwt"
    idx_fwd_2 = idx_prefix + ".2.ebwt"
    idx_rev_1 = idx_prefix + ".rev.1.ebwt"
    idx_rev_2 = idx_prefix + ".rev.2.ebwt"
    
    if os.path.exists(idx_fwd_1) and \
       os.path.exists(idx_fwd_2) and \
       os.path.exists(idx_rev_1) and \
       os.path.exists(idx_rev_2):
        return 
    else:
        bowtie_idx_env_var = os.environ.get("BOWTIE_INDEXES")
        if bowtie_idx_env_var == None:
            print >> sys.stderr, "Error: Could not find Bowtie index files " + idx_prefix + ".*"
            exit(1)
        idx_prefix = bowtie_idx_env_var + idx_prefix 
        idx_fwd_1 = idx_prefix + ".1.ebwt"
        idx_fwd_2 = idx_prefix + ".2.ebwt"
        idx_rev_1 = idx_prefix + ".rev.1.ebwt"
        idx_rev_2 = idx_prefix + ".rev.2.ebwt"
        
        if os.path.exists(idx_fwd_1) and \
           os.path.exists(idx_fwd_2) and \
           os.path.exists(idx_rev_1) and \
           os.path.exists(idx_rev_2):
            return 
        else:
            print >> sys.stderr, "Error: Could not find Bowtie index files " + idx_prefix + ".*"
            exit(1)

def bowtie_idx_to_fa(idx_prefix):
    idx_name = idx_prefix.split('/')[-1]
    print >> sys.stderr, "[%s] Reconstituting reference FASTA file from Bowtie index" % (right_now())
    
    try:    
        tmp_fasta_file_name = output_dir + idx_name + ".fa"
        tmp_fasta_file = open(tmp_fasta_file_name, "w")

        inspect_log = open(logging_dir + "bowtie_inspect.log", "w")

        inspect_cmd = ["bowtie-inspect",
                       idx_prefix]
        #print >> sys.stderr, "Executing: " + " ".join(inspect_cmd) + " > " + tmp_fasta_file_name   
        ret = subprocess.call(inspect_cmd, 
                              stdout=tmp_fasta_file,
                              stderr=inspect_log)

        # Bowtie reported an error
        if ret != 0:
           print >> sys.stderr, fail_str, "Error: bowtie-inspect returned an error"
           exit(1)
           
    # Bowtie not found
    except OSError, o:
        if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
            print >> sys.stderr, fail_str, "Error: bowtie-inspect not found on this system.  Did you forget to include it in your PATH?"
  
    return tmp_fasta_file_name

def check_fasta(idx_prefix):
    print >> sys.stderr, "[%s] Checking for reference FASTA file" % right_now()
    idx_fasta = idx_prefix + ".fa"
    if os.path.exists(idx_fasta):
        return idx_fasta
    else:
        idx_name = idx_prefix.split('/')[-1]
        bowtie_idx_env_var = os.environ.get("BOWTIE_INDEXES")
        if bowtie_idx_env_var != None:
            idx_fasta = bowtie_idx_env_var + idx_prefix + ".fa" 
            if os.path.exists(idx_fasta):
                return idx_fasta
        
        print >> sys.stderr, "\tWarning: Could not find FASTA file " + idx_fasta
        idx_fa = bowtie_idx_to_fa(idx_prefix)
        return idx_fa
        #print >> sys.stderr, "Error: Could not find Maq binary fasta file " + idx_bfa
        #exit(1)
    
def check_index(idx_prefix):
    check_bowtie_index(idx_prefix)
    ref_fasta_file = check_fasta(idx_prefix)
    return (ref_fasta_file, None)

def get_bowtie_version():
    try:
        # Launch Bowtie to capture its version info
        proc = subprocess.Popen(['bowtie', '--version'],stdout=subprocess.PIPE)
        stdout_value = proc.communicate()[0]
        bowtie_version = None
        bowtie_out = repr(stdout_value)

        # Find the version identifier
        version_str = "bowtie version "
        ver_str_idx = bowtie_out.find(version_str)
        if ver_str_idx != -1:
            nl = bowtie_out.find("\\n", ver_str_idx)
            version_val = bowtie_out[ver_str_idx + len(version_str):nl]
            bowtie_version = [int(x) for x in version_val.split('.')]
        if len(bowtie_version) == 3:
            bowtie_version.append(0)
        
        return bowtie_version
    except OSError, o:
       if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
           print >> sys.stderr, fail_str, "Error: bowtie not found on this system"
       exit(1)

def check_bowtie():
    print >> sys.stderr, "[%s] Checking for Bowtie" % right_now()
    bowtie_version = get_bowtie_version()
    if bowtie_version == None:
        print >> sys.stderr, "Error: Bowtie not found on this system"
        exit(1)
    elif bowtie_version[1] < 10:
        print >> sys.stderr, "Error: TopHat requires Bowtie 0.10.0 or later"
        exit(1)
    print >> sys.stderr, "\tBowtie version:\t\t %s" % ".".join([str(x) for x in bowtie_version])
        
        
def check_reads(params, reads_files):
    print >> sys.stderr, "[%s] Checking reads" % right_now()
    bowtie_version = get_bowtie_version()
    
    seed_len = params.seed_length
    format = params.reads_format
    
    observed_formats = set([])
    observed_scales = set([])
    min_seed_len = 99999
    max_qual = -1
    files = reads_files.split(',')
    for f_name in files:
        try:
            f = open(f_name)
        except IOError:
            print >> sys.stderr, "Error: could not open file", f_name
            exit(1)
            
        first_line = f.readline()
        if first_line[0] == "@":
            format = "fastq"
        elif first_line[0] == ">":
            format = "fasta"
        else:
            print >> sys.stderr, "Error: file %s does not appear to be a valid FASTA or FASTQ file" % f_name
        observed_formats.add(format)
        f.seek(0)
        line_num = 0
        if format == "fastq":
            for line in f:
                if line_num % 4 == 1:
                    seq_len = len(line) - 1
                    if seq_len < 20:
                        print >> sys.stderr, "Warning: found a read < 20bp in", f_name
                    else:
                        min_seed_len = min(seq_len, min_seed_len)                
                elif line_num % 4 == 3:
                    max_line_qual = max([ord(x) for x in list(line.strip())])
                    max_qual = max(max_line_qual, max_qual)
                line_num += 1
            if max_qual > 90:
                quality_scale = "--solexa_scale"

        elif format == "fasta":
            for line in f:
                if line_num % 2 == 1:
                    seq_len = len(line) - 1
                    if seq_len < 20:
                        print >> sys.stderr, "Warning: found a read < 20bp in", f_name
                    else:
                        min_seed_len = min(seq_len, min_seed_len)
                line_num += 1
            
    if len(observed_formats) > 1:
        print >> sys.stderr, "Error: TopHat requires all reads be either FASTQ or FASTA.  Mixing formats is not supported."
        exit(1)
    
    if seed_len != None:
        seed_len = min(seed_len, min_seed_len)
    else:
        seed_len = min_seed_len
        
    print >> sys.stderr, "\tseed length:\t %dbp" % seed_len
    print >> sys.stderr, "\tformat:\t\t %s" % format
    if format == "fastq":
        print >> sys.stderr, "\tquality scale:\t %s" % params.quality_scale
    
    #print seed_len, format, solexa_scale
    return TopHatParams.ReadParams(params.quality_scale, 
                                   seed_len, 
                                   format, 
                                   params.mate_inner_dist, 
                                   params.mate_inner_dist_std_dev)

def formatTD(td):
  hours = td.seconds // 3600
  minutes = (td.seconds % 3600) // 60
  seconds = td.seconds % 60
  return '%02d:%02d:%02d' % (hours, minutes, seconds) 

def prep_reads(reads_format, reads_list, output_name):    
    #filter_cmd = ["prep_reads"]
    if reads_format == "fasta":
        reads_format = "-f"
    else:
        reads_format = "-q"
        
    reads_suffix = ".fq"
    kept_reads_filename = output_dir + output_name + reads_suffix
    
    if os.path.exists(kept_reads_filename):
        os.remove(kept_reads_filename)
    kept_reads = open(kept_reads_filename, "a")
    
    filter_log = open(logging_dir + "prep_reads.log", "w")
    
    filter_cmd = [bin_dir + "prep_reads",
                  "-Q",
                  reads_format,
                  reads_list]   
    #print "\t executing: `%s'" % " ".join(filter_cmd)    
    # files = reads_list.split(',')
    # for reads_file in files:
    try:       
        print >> run_log, " ".join(filter_cmd)
        ret = subprocess.call(filter_cmd, 
                              stdout=kept_reads,
                              stderr=filter_log)
                              # Bowtie reported an error
        if ret != 0:
            print >> sys.stderr, fail_str, "Error: could not execute prep_reads"
            exit(1)
    # prep_reads not found
    except OSError, o:
        if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
            print >> sys.stderr, fail_str, "Error: prep_reads not found on this system.  Did you forget to include it in your PATH?"
        exit(1)
        
    return kept_reads_filename

def bowtie(params,
           bwt_idx_prefix,
           reads_list,
           reads_format,
           mapped_reads,
           unmapped_reads,
           reads_for_ordering=None,
           phred_thresh=70):
    start_time = datetime.now()
    bwt_idx_name = bwt_idx_prefix.split('/')[-1]
    print >> sys.stderr, "[%s] Mapping reads against %s with Bowtie" % (start_time.strftime("%c"), bwt_idx_name)
    
    # Setup Bowtie output redirects
    #bwt_map = output_dir + mapped_reads
    bwt_map = tmp_name()
    bwt_log = open(logging_dir + "unspliced_bwt.log", "w")
    
    # Launch Bowtie
    try:    
        bowtie_cmd = ["bowtie"]
        
        bowtie_cmd += [params.read_params.quality_scale]
        
        if reads_format == "fastq":
            bowtie_cmd += ["-q"]
        elif reads_format == "fasta":
            bowtie_cmd += ["-f"]
            
        if unmapped_reads != None:
            unmapped_reads_fasta_name = unmapped_reads
            bowtie_cmd += ["--un", unmapped_reads_fasta_name,
                           "--max", "/dev/null"]
        else:
            unmapped_reads_fasta_name = None
        
        bowtie_cmd += ["-v", str(params.segment_mismatches),
                         "-p", str(params.system_params.bowtie_threads),
                         "-k", str(params.max_hits),
                         "-m", str(params.max_hits),
                         bwt_idx_prefix, 
                         reads_list]
        
        bowtie_proc = subprocess.Popen(bowtie_cmd, stdout=subprocess.PIPE, stderr=open(bwt_map+".log","w"))
        
        fix_map_cmd = [bin_dir + "fix_map_ordering"]
        
        if reads_format == "fastq":
            fix_map_cmd += ["--fastq"]
        elif reads_format == "fasta":
            fix_map_cmd += ["--fasta"]
        
        if reads_for_ordering == None:
            reads_for_ordering = reads_list
        
        fix_map_cmd.extend([reads_for_ordering, "-"])
             
        print >> run_log, " ".join(bowtie_cmd),"|", " ".join(fix_map_cmd), ">", bwt_map
        
        bwt_map = mapped_reads 
        fix_order_proc = subprocess.Popen(fix_map_cmd, 
                                          stdin=bowtie_proc.stdout,
                                          stdout=open(mapped_reads, "w"))    
        
        # wait for the whole pipe to finish
        fix_order_proc.communicate()   
            
    # Bowtie not found
    except OSError, o:
        if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
            print >> sys.stderr, fail_str, "Error: Bowtie not found on this system.  Did you forget to include it in your PATH?"
        exit(1)
            
    # Success    
    finish_time = datetime.now()
    duration = finish_time - start_time
    #print >> sys.stderr, "\t\t\t[%s elapsed]" %  formatTD(duration)
    return (bwt_map, unmapped_reads_fasta_name)

def tmp_name():
    tmp_root = output_dir + "tmp/"
    if os.path.exists(tmp_root):
        pass
    else:        
        os.mkdir(tmp_root)
    return tmp_root + os.tmpnam().split('/')[-1] 

def get_gff_juncs(gff_annotation):
    print >> sys.stderr, "[%s] Reading known junctions from GFF file" % (right_now())
    gff_juncs_log = open(logging_dir + "gff_juncs.log", "w")
    
    gff_prefix = gff_annotation.split('/')[-1].split('.')[0]
    
    gff_juncs_out_name  = output_dir + gff_prefix + ".juncs"
    gff_juncs_out = open(gff_juncs_out_name, "w")
    
    gff_juncs_cmd = [bin_dir + "gff_juncs", gff_annotation]            
    try:    
        print >> run_log, " ".join(gff_juncs_cmd)
        retcode = subprocess.call(gff_juncs_cmd, 
                                  stderr=gff_juncs_log,
                                  stdout=gff_juncs_out)
        # cvg_islands returned an error
        if retcode == 1:
            print >> sys.stderr, "\tWarning: TopHat did not find any junctions in GFF file"
            return (False, gff_juncs_out_name) 
        elif retcode < 0:
            print >> sys.stderr, fail_str, "Error: GFF junction extraction failed with err =", retcode
            exit(1)
    # cvg_islands not found
    except OSError, o:
       if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
           print >> sys.stderr, fail_str, "Error: gff_juncs not found on this system"
       exit(1)
    return (True, gff_juncs_out_name)


def build_juncs_bwt_index(external_splice_prefix):
    print >> sys.stderr, "[%s] Indexing splices" % (right_now())
    bowtie_build_log = open(logging_dir + "bowtie_build.log", "w")
    
    #user_splices_out_prefix  = output_dir + "user_splices_idx"
    
    bowtie_build_cmd = ["bowtie-build", 
                        external_splice_prefix + ".fa",
                        external_splice_prefix]            
    try:    
        print >> run_log, " ".join(bowtie_build_cmd)
        retcode = subprocess.call(bowtie_build_cmd, 
                                 stdout=bowtie_build_log)
       
        if retcode != 0:
            print >> sys.stderr, fail_str, "Error: Splice sequence indexing failed with err =", retcode
            exit(1)
    except OSError, o:
        if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
            print >> sys.stderr, fail_str, "Error: bowtie-build not found on this system"
        exit(1)
    return external_splice_prefix
    
def build_juncs_index(min_anchor_length, 
                      read_length,
                      juncs_prefix, 
                      external_juncs,  
                      reference_fasta):
    print >> sys.stderr, "[%s] Retrieving sequences for splices" % (right_now())
    
    juncs_file_list = ",".join(external_juncs)
    juncs_db_log = open(logging_dir + "juncs_db.log", "w")
    
    external_splices_out_prefix  = output_dir + juncs_prefix
    external_splices_out_name = external_splices_out_prefix + ".fa"
    
    external_splices_out = open(external_splices_out_name, "w")
    
    juncs_db_cmd = [bin_dir + "juncs_db", 
                    str(min_anchor_length),
                    str(read_length),
                    juncs_file_list,
                    reference_fasta]            
    try:    
        print >> run_log, " ".join(juncs_db_cmd)
        retcode = subprocess.call(juncs_db_cmd, 
                                 stderr=juncs_db_log,
                                 stdout=external_splices_out)
       
        if retcode != 0:
            print >> sys.stderr, fail_str, "Error: Splice sequence retrieval failed with err =", retcode
            exit(1)
    # cvg_islands not found
    except OSError, o:
       if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
           print >> sys.stderr, fail_str, "Error: juncs_db not found on this system"
       exit(1)
       
    external_splices_out_prefix = build_juncs_bwt_index(external_splices_out_prefix)
       
    return external_splices_out_prefix

def compile_reports(params, left_maps, left_reads, right_maps, right_reads, gff_annotation):
    print >> sys.stderr, "[%s] Reporting output tracks" % right_now()
    
    left_maps = [x for x in left_maps if (os.path.exists(x) and os.path.getsize(x) > 0)]
    left_maps = ','.join(left_maps)
    
    if len(right_maps) > 0:
        right_maps = [x for x in right_maps if (os.path.exists(x) and os.path.getsize(x) > 0)]
        right_maps = ','.join(right_maps)
    
    report_log = open(logging_dir + "reports.log", "w")
    junctions = "junctions.bed"
    coverage =  "coverage.wig"
    accepted_hits = "accepted_hits.sam"
    report_cmd = [bin_dir + "tophat_reports"]
    report_cmd.extend(params.cmd())
        
    report_cmd.extend([junctions,
                       accepted_hits,
                       left_maps,
                       left_reads])
    if len(right_maps) > 0 and right_reads != None:
        report_cmd.append(right_maps)
        report_cmd.append(right_reads)
                    
    try: 
        print >> run_log, " ".join(report_cmd)   
        retcode = subprocess.call(report_cmd, 
                                  stderr=report_log)
       
        # spanning_reads returned an error 
        if retcode != 0:
            print >> sys.stderr, fail_str, "Error: Report generation failed with err =", retcode
            exit(1)
            
        sorted_map = tmp_name()
        sort_cmd =["sort",
                    "-k",
                    "3,3", 
                    "-k", 
                    "4,4n",
                    "--temporary-directory="+tmp_dir,
                    output_dir + accepted_hits]
                    
        print >> run_log, " ".join(sort_cmd), ">",sorted_map
        
        subprocess.call(sort_cmd, 
                        stdout=open(sorted_map, "w"),
                        stderr=report_log)
        os.rename(sorted_map, output_dir + accepted_hits) 
        
        wig_cmd = ["wiggles", output_dir + accepted_hits, output_dir + coverage]
        print >> run_log, " ".join(wig_cmd)
        subprocess.call(wig_cmd,
                        stderr=open("/dev/null"))
                        
    # cvg_islands not found
    except OSError, o:
        if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
            print >> sys.stderr, fail_str, "Error: tophat_reports not found on this system"
        exit(1)
    return (coverage, junctions)

def split_reads(reads_filename, 
                prefix, 
                fasta, 
                read_length, 
                segment_length):
    reads_file = open(reads_filename)
    output_files = []
    num_segments = read_length / segment_length
    
    if num_segments == 1:
        segment_length = read_length
    
    if fasta == True:
        extension = ".fa"
    else:
        extension = ".fq"
    print >> sys.stderr, "\tSplitting reads into %d segments" % (num_segments)
    def open_output_files(prefix, num_files, output_files, extension):
        i = 1
        while i <= num_files:
            output_files.append(open(prefix + ("_seg%d" % i) + extension, "w"))
            i += 1
    def split_record(read_name, read_seq, read_qual, output_files, offsets):
        seg_num = 0
        last_seq_offset = 0
        while seg_num < len(offsets) - 1:
            f = output_files[seg_num]
            #print last_seq_offset, offsets[seg_num + 1]
            seg_seq = read_seq[last_seq_offset:offsets[seg_num + 1]]
            if fasta == False: #FASTQ
                seg_qual = read_qual[last_seq_offset:offsets[seg_num + 1]]
                print >> f, "%s|%d:%d" % (read_name,last_seq_offset,seg_num)
                print >> f, seg_seq
                print >> f, "+"
                print >> f, seg_qual
            else:
                print >> f, "%s|%d:%d" % (read_name, last_seq_offset,seg_num)
                print >> f, seg_seq
            seg_num += 1
            last_seq_offset = offsets[seg_num]
    offsets = [segment_length * i for i in range(0,num_segments + 1)]
    
    # Bowtie's minimum read length here is 20bp, so if the last segment
    # is between 20 and segment_length bp long, go ahead and write it out
    #print >> sys.stderr, read_length % segment_length, num_segments
    if read_length % segment_length >= 20:
        offsets.append(segment_length * (num_segments + 1))
        num_segments += 1
    open_output_files(prefix, num_segments, output_files, extension)
    if fasta:
        line_state = 0
        read_name = ""
        read_seq = ""
        for line in reads_file:
            if line_state == 0:
                read_name = line.strip()
            else:
                read_seq = line.strip()
                split_record(read_name, read_seq, None, output_files, offsets)
            line_state += 1
            line_state %= 2
    else:
        line_state = 0
        read_name = ""
        read_seq = ""
        read_qual = ""
        for line in reads_file:
            if line.strip() == "":
                continue
            #print >> sys.stderr, line,
            if line_state == 0:
                read_name = line.strip()
            elif line_state == 1:
                read_seq = line.strip()
            elif line_state == 2:
                line = line.strip()
            else:
                read_quals = line.strip()
                split_record(read_name, read_seq, read_quals, output_files, offsets)
            line_state += 1
            line_state %= 4
    for f in output_files:
        f.close()
    return [o.name for o in output_files]

def junctions_from_closures(params,
                            left_maps, 
                            right_maps, 
                            ref_fasta):
    #print >> sys.stderr, "[%s] " % right_now()
    print >> sys.stderr, "[%s] Searching for junctions via mate-pair closures" % right_now()
    #maps = [x for x in seg_maps if (os.path.exists(x) and os.path.getsize(x) > 0)]
    #if len(maps) == 0:
    #    return None
    #print >> sys.stderr, left_maps
    slash = left_maps.rfind('/')
    juncs_out = ""
    if slash != -1:
        juncs_out += left_maps[:slash + 1]
    juncs_out += "closure.juncs"

    juncs_log = open(logging_dir + "closure.log", "w")
    juncs_cmd = [bin_dir + "closure_juncs"]

    # if reads_format == "fastq":
    #     align_cmd += ["-q"]
    # elif reads_format == "fasta":
    #     align_cmd += ["-f"]

    juncs_cmd.extend(params.cmd())
    juncs_cmd.extend([juncs_out,
                      ref_fasta,
                      left_maps,
                      right_maps])            
    try:
        print >> run_log, ' '.join(juncs_cmd)
        retcode = subprocess.call(juncs_cmd, 
                                 stderr=juncs_log)

        # spanning_reads returned an error 
        if retcode != 0:
           print >> sys.stderr, fail_str, "Error: closure-based junction search failed with err =", retcode
           exit(1)
    # cvg_islands not found
    except OSError, o:
        if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
           print >> sys.stderr, fail_str, "Error: closure_juncs not found on this system"
        exit(1)
    return [juncs_out]

def junctions_from_segments(params,
                            left_reads, 
                            left_seg_maps, 
                            right_reads,
                            right_seg_maps,
                            unmapped_reads, 
                            reads_format, 
                            ref_fasta):
    #print >> sys.stderr, "[%s] " % right_now()
    print >> sys.stderr, "[%s] Searching for junctions via segment mapping" % right_now()
    #maps = [x for x in seg_maps if (os.path.exists(x) and os.path.getsize(x) > 0)]
    #if len(maps) == 0:
    #    return None
    slash = left_seg_maps[0].rfind('/')
    juncs_out = ""
    if slash != -1:
        juncs_out += left_seg_maps[0][:slash + 1]
    juncs_out += "segment.juncs"
    
    left_maps = ','.join(left_seg_maps)
    align_log = open(logging_dir + "segment_juncs.log", "w")
    align_cmd = [bin_dir + "segment_juncs"]
    
    align_cmd.extend(params.cmd())
    
    align_cmd.extend(["--ium-reads", ",".join(unmapped_reads),
                      ref_fasta,
                      juncs_out,
                      left_reads,
                      left_maps])
    if right_seg_maps != None:
        right_maps = ','.join(right_seg_maps)
        align_cmd.extend([right_reads, right_maps])            
    try:
        print >> run_log, " ".join(align_cmd)
        retcode = subprocess.call(align_cmd, 
                                 stderr=align_log)

        # spanning_reads returned an error 
        if retcode != 0:
           print >> sys.stderr, fail_str, "Error: segment-based junction search failed with err =",retcode
           exit(1)
    # cvg_islands not found
    except OSError, o:
        if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
           print >> sys.stderr, fail_str, "Error: segment_juncs not found on this system"
        exit(1)

    return [juncs_out]

def join_mapped_segments(params,
                         reads,
                         ref_fasta,
                         possible_juncs,
                         contig_seg_maps,
                         spliced_seg_maps,
                         alignments_out_name):
    print >> sys.stderr, "[%s] Joining segment hits" % right_now()
    contig_seg_maps = ','.join(contig_seg_maps)
    possible_juncs = ','.join(possible_juncs)
    
    align_log = open(logging_dir + "long_spanning_reads.log", "w")
    align_cmd = [bin_dir + "long_spanning_reads"]
    
    # if params.read_params.reads_format == "fastq":
    #     align_cmd += ["-q"]
    # elif params.read_params.reads_format == "fasta":
    #     align_cmd += ["-f"]
    
    alignments_out = open(alignments_out_name, "w")
    
    align_cmd.extend(params.cmd())
    
    align_cmd.extend([ reads,
                       possible_juncs,
                       contig_seg_maps])
    if spliced_seg_maps != None:
        spliced_seg_maps = ','.join(spliced_seg_maps)
        align_cmd.append(spliced_seg_maps)
                
    try:    
        print >> run_log, " ".join(align_cmd),">", alignments_out_name
        retcode = subprocess.call(align_cmd, 
                                  stderr=align_log,
                                  stdout=alignments_out)

        # spanning_reads returned an error 
        if retcode != 0:
            print >> sys.stderr, fail_str, "Error: Segment join failed with err =", retcode
            exit(1)
     # cvg_islands not found
    except OSError, o:
        if o.errno == errno.ENOTDIR or o.errno == errno.ENOENT:
            print >> sys.stderr, fail_str, "Error: long_spanning_reads not found on this system"
        exit(1)
      
    
def spliced_alignment(params,
                      bwt_idx_prefix,
                      ref_fasta,
                      read_len,
                      segment_len,
                      left_reads,
                      right_reads,
                      user_supplied_junctions):
    
    possible_juncs = []
    possible_juncs.extend(user_supplied_junctions)
    
    left_maps = []
    right_maps = []
    maps = { left_reads : [], right_reads : [] }
    #single_segments = False
    
    class Maps:
        def __init__(self, 
                     unspliced_bwt,
                     unspliced_sam,
                     seg_maps, 
                     unmapped_segs, 
                     segs):
            self.unspliced_bwt = unspliced_bwt
            self.unspliced_sam = unspliced_sam
            self.seg_maps = seg_maps
            self.unmapped_segs = unmapped_segs
            self.segs = segs
                  
    for reads in [left_reads, right_reads]:
        if reads == None:
            continue
        slash = reads.rfind("/")
        extension = reads.rfind(".")
        if extension != -1:
            prefix = reads[slash + 1:extension]
        else:
            prefix = reads      
          
        extension = reads.rfind(".")
        assert extension != -1
        tmp = reads.rfind("/")

        unspliced_out = tmp_dir + reads[tmp:extension] + ".bwtout"  
        unmapped_unspliced = tmp_dir + reads[tmp:extension] + "_missing.fq"
        num_segs = read_len / segment_len
        phred_thresh = 70 * num_segs
        (unspliced_map, unmapped) = bowtie(params,
                                           bwt_idx_prefix, 
                                           reads,
                                           "fastq",
                                           unspliced_out,
                                           unmapped_unspliced,
                                           reads,
                                           phred_thresh)  
        unspliced_sam = tmp_name()
        join_mapped_segments(params,
                             reads,
                             ref_fasta,
                             ["/dev/null"],
                             [unspliced_map],
                             [],
                             unspliced_sam)
            
        if num_segs == 1:
            segment_len = read_len
        elif num_segs >= 3:
            if params.closure_search != True:
                params.closure_search = False
            if params.coverage_search != True:
                params.coverage_search = False
            if params.butterfly_search != True:
                params.butterfly_search = False
        if params.closure_search != True:
            params.closure_search = False
                                                       

        
        #print spliced_params.read_params.seed_len
        seg_maps = []
        unmapped_segs = []
        segs = []
        if num_segs > 1:
            read_segments = split_reads(unmapped_unspliced,
                                        output_dir + "/tmp/" + prefix, 
                                        False,
                                        read_len, 
                                        segment_len)
            for seg in read_segments:
                extension = seg.rfind(".")
                assert extension != -1
                tmp = seg.rfind("/")
                assert tmp != -1
                seg_out =  tmp_dir + seg[tmp:extension] + ".bwtout"
                unmapped_seg = tmp_dir + seg[tmp:extension] + "_missing.fq"
                (seg_map, unmapped) = bowtie(params,
                                             bwt_idx_prefix, 
                                             seg,
                                             "fastq",
                                             seg_out,
                                             unmapped_seg,
                                             seg)
                seg_maps.append(seg_map)
                unmapped_segs.append(unmapped)
                segs.append(seg)
            
            maps[reads] = Maps(unspliced_map, unspliced_sam, seg_maps, unmapped_segs, segs) 
        else:
            read_segments = [reads]
            maps[reads] = Maps(unspliced_map, unspliced_sam, [unspliced_map], [unmapped_unspliced], read_segments)
    
    if params.find_novel_juncs:    
        left_seg_maps = maps[left_reads].seg_maps
        unmapped_reads = maps[left_reads].unmapped_segs
        if right_reads != None:
            right_seg_maps = maps[right_reads].seg_maps
            unmapped_reads.extend(maps[right_reads].unmapped_segs)
        else:
            right_seg_maps = None
        juncs = junctions_from_segments(params, 
                                        left_reads, 
                                        left_seg_maps, 
                                        right_reads,
                                        right_seg_maps,
                                        unmapped_reads,
                                        "fastq", 
                                        ref_fasta)
        if os.path.getsize(juncs[0]) != 0:
            possible_juncs.extend(juncs)
    
        if params.closure_search and left_reads != None and right_reads != None:
            juncs = junctions_from_closures(params,
                                            maps[left_reads].seg_maps[-1], 
                                            maps[right_reads].seg_maps[-1],
                                            ref_fasta)
            if os.path.getsize(juncs[0]) != 0:
                possible_juncs.extend(juncs)         
    
    if len(possible_juncs) == 0:
        spliced_seg_maps = None
        junc_idx_prefix = None
        print >> sys.stderr, "Warning: junction database is empty!"
    else:  
        junc_idx_prefix = "tmp/segment_juncs"
        build_juncs_index(3,
                          segment_len,
                          junc_idx_prefix, 
                          possible_juncs,
                          ref_fasta)
                          
    for reads in [left_reads, right_reads]:
        spliced_seg_maps = []
        if reads == None or reads not in maps:
            continue
        
        if junc_idx_prefix != None:
            i = 0    
            for seg in maps[reads].segs:
                extension = seg.rfind(".")
                assert extension != -1
                tmp = seg.rfind("/")
                assert tmp != -1
            
                ordering = maps[reads].segs[i]
                seg_out = tmp_dir + seg[tmp:extension] + "_to_spliced.bwtout"
                #unmapped_seg = seg[tmp:extension] + "_missing.fa"
                (seg_map, unmapped) = bowtie(params,
                                             output_dir + junc_idx_prefix, 
                                             seg,
                                             "fastq",
                                             seg_out,
                                             None,
                                             ordering)
                spliced_seg_maps.append(seg_map)
                i += 1


        mapped_reads = reads +".candidate_hits.sam"
        join_mapped_segments(params,
                             reads,
                             ref_fasta,
                             possible_juncs,
                             maps[reads].seg_maps,
                             spliced_seg_maps,
                             mapped_reads)
        if num_segs > 1:
            merged_map = tmp_name()
            merge_sort_cmd =["sort",
                             "-k 1,1n",
                              "--temporary-directory="+tmp_dir,
                              maps[reads].unspliced_sam, 
                              mapped_reads]
            print >> run_log, " ".join(merge_sort_cmd), ">", merged_map
            subprocess.call(merge_sort_cmd,
                             stdout=open(merged_map,"w"))  
        else:
            merged_map = mapped_reads
        
        if params.system_params.keep_tmp == False:
            os.remove(maps[reads].unspliced_bwt)
            os.remove(maps[reads].unspliced_sam)            
            os.rename(merged_map, mapped_reads)
                          
        maps[reads] = [mapped_reads]
    return maps

def get_version():
   return "1.0.12"

def main(argv=None):
    warnings.filterwarnings("ignore", "tmpnam is a potential security risk")
    
    # Initialize default parameter values
    params = TopHatParams()
    
    try:
        if argv is None:
            argv = sys.argv
            args = params.parse_options(argv)
            params.check()
        
        bwt_idx_prefix = args[0]
        left_reads_list = args[1]
        if len(args) > 2:
            if params.read_params.mate_inner_dist == None:
                print >> sys.stderr, "Error: you must set the mean inner distance between mates with -r"
                exit(1)
            right_reads_list = args[2]
        else:
            right_reads_list = None
            
        print >> sys.stderr
        print >> sys.stderr, "[%s] Beginning TopHat run (v%s)" % (right_now(), get_version())
        print >> sys.stderr, "-----------------------------------------------" 
        
        start_time = datetime.now()
        prepare_output_dir()
        
        global run_log
        run_log = open(logging_dir + "run.log", "w", 0)
        print >> run_log, " ".join(argv)
        
        
        # Validate all the input files, check all prereqs before committing 
        # to the run
        (ref_fasta, idx_bfa) = check_index(bwt_idx_prefix)
        #use_long_maq_maps = check_maq()
        check_bowtie()
        
        if params.skip_check_reads == False:
            # TODO: check right reads as well ?
            params.read_params = check_reads(params.read_params, left_reads_list)
            
        user_supplied_juncs = []
        if params.gff_annotation != None and params.find_GFF_juncs == True:
            (found_juncs, gff_juncs) = get_gff_juncs(params.gff_annotation)
            if found_juncs == True:
                user_supplied_juncs.append(gff_juncs)
        if params.raw_junctions != None:
            user_supplied_juncs.append(params.raw_junctions)
                
        # Now start the time consuming stuff
        left_kept_reads = prep_reads(params.read_params.reads_format,
                                     left_reads_list,
                                     "left_kept_reads")
                                     
        if right_reads_list != None:
            right_kept_reads = prep_reads(params.read_params.reads_format,
                                          right_reads_list,
                                          "right_kept_reads")
        else:
            right_kept_reads = None
            
        spliced_reads = []
        
        #if params.read_params.seed_length / segment_length > 1:
        
        mapping = spliced_alignment(params, 
                                    bwt_idx_prefix,
                                    ref_fasta,
                                    params.read_params.seed_length,
                                    params.segment_length,
                                    left_kept_reads,
                                    right_kept_reads,
                                    user_supplied_juncs)
                                    
        left_maps = mapping[left_kept_reads]
        #left_unmapped_reads = mapping[1]
        right_maps = mapping[right_kept_reads]
        #right_unmapped_reads = mapping[3]
            
        compile_reports(params,
                        left_maps,
                        left_kept_reads,
                        right_maps,
                        right_kept_reads,
                        params.gff_annotation)
                        
        if params.system_params.keep_tmp == False:
            for m in left_maps:
                os.remove(m)
            if left_kept_reads != None:
                os.remove(left_kept_reads)
            for m in right_maps:
                os.remove(m)
            if right_kept_reads != None:
                os.remove(right_kept_reads)
            tmp_files = os.listdir(tmp_dir)
            for t in tmp_files:
                os.remove(tmp_dir+t)
            os.rmdir(tmp_dir)
        
        finish_time = datetime.now()
        duration = finish_time - start_time
        print >> sys.stderr,"-----------------------------------------------"
        print >> sys.stderr, "Run complete [%s elapsed]" %  formatTD(duration)
        
    except Usage, err:
        print >> sys.stderr, sys.argv[0].split("/")[-1] + ": " + str(err.msg)
        print >> sys.stderr, "    for detailed help see http://tophat.cbcb.umd.edu/manual.html"
        return 2


if __name__ == "__main__":
    sys.exit(main())
