/*
 *  tophat_reports.cpp
 *  TopHat
 *
 *  Created by Cole Trapnell on 11/20/08.
 *  Copyright 2008 Cole Trapnell. All rights reserved.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cassert>
#include <cstdio>
#include <cstring>
#include <vector>
#include <string>
#include <map>
#include <algorithm>
#include <set>
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <seqan/sequence.h>
#include <seqan/find.h>
#include <seqan/file.h>
#include <getopt.h>

#include "common.h"
#include "bwt_map.h"
#include "junctions.h"
#include "fragments.h"
#include "wiggles.h"
#include "tokenize.h"
#include "genes.h"
#include "reads.h"

#include "inserts.h"

using namespace std;
using namespace seqan;
using std::set;

void fragment_best_alignments(const HitsForRead& hits_for_read,
							  const JunctionSet& junctions,
							  FragmentAlignmentGrade& best_grade,
                              HitsForRead& best_hits)
{
	const vector<BowtieHit>& hits = hits_for_read.hits;
	
	for (size_t i = 0; i < hits.size(); ++i)
	{
		FragmentAlignmentGrade g(hits[i]);
		// Is the new status better than the current best one?
		if (best_grade < g)
		{
			best_hits.hits.clear();
			best_grade = g;
			best_hits.hits.push_back(hits[i]);
		}
		else if (!(g < best_grade)) // is it just as good?
		{
			best_grade.num_alignments++;
			best_hits.hits.push_back(hits[i]);
		}
	}
}

void insert_best_alignments(const HitsForRead& left_hits,
							const HitsForRead& right_hits,
							const JunctionSet& junctions,
							InsertAlignmentGrade& best_grade,
							HitsForRead& left_best_hits,
							HitsForRead& right_best_hits)
{
	// max mate inner distance (genomic)
	int min_mate_inner_dist = inner_dist_mean - inner_dist_std_dev;
	if (max_mate_inner_dist == -1)
	{
		max_mate_inner_dist = inner_dist_mean + inner_dist_std_dev;
	}
	
	const vector<BowtieHit>& left = left_hits.hits;
	const vector<BowtieHit>& right = right_hits.hits;
	
	for (size_t i = 0; i < left.size(); ++i)
	{
		for (size_t j = 0; j < right.size(); ++j)
		{
		
			const BowtieHit& lh = left[i];
			const BowtieHit& rh = right[j];
			
			InsertAlignmentGrade g(lh, rh, min_mate_inner_dist, max_mate_inner_dist);

			// Is the new status better than the current best one?
			if (best_grade < g)
			{
				left_best_hits.hits.clear();
				right_best_hits.hits.clear();
				
				best_grade = g;
				left_best_hits.hits.push_back(lh);
				right_best_hits.hits.push_back(rh);
			}
			else if (!(g < best_grade))
			{
				left_best_hits.hits.push_back(lh);
				right_best_hits.hits.push_back(rh);
				best_grade.num_alignments++;
			}
		}
	}
}

enum FragmentType {FRAG_UNPAIRED, FRAG_LEFT, FRAG_RIGHT};

bool rewrite_sam_hit(const BowtieHit& bh,
					 const char* bwt_buf,
					 char* rebuf, 
					 char* read_alt_name,
					 const FragmentAlignmentGrade& grade,
					 FragmentType insert_side)
{
	// Rewrite this hit, filling in the alt name, mate mapping
	// and setting the pair flag
	vector<string> sam_toks;
	tokenize(bwt_buf, "\t", sam_toks);
	char* slash_pos = strrchr(read_alt_name,'/');
	if (slash_pos)
	{
		*slash_pos = 0;
	}
	
	*rebuf = 0;
	
	for (size_t t = 0; t < sam_toks.size(); ++t)
	{
		switch(t)
		{
			case 0: //QNAME
			{
				sam_toks[t] = read_alt_name;
				break;
			}
			
			case 1:
			{
				// SAM FLAG
				if (insert_side != FRAG_UNPAIRED)
				{
					int flag = 0;
					// mark this guys as a singleton mate
					flag |= 0x0001;
					if (insert_side == FRAG_LEFT)
						flag |= 0x0040;
					else if (insert_side == FRAG_RIGHT)
						flag |= 0x0080;
					flag |= 0x0008;
					
					char flag_buf[64];
					sprintf(flag_buf, "%d", flag);
					sam_toks[t] = flag_buf;
				}
				break;
			}
				
			case 4: //MAPQ
			{
				int mapQ;
				if (grade.num_alignments > 1)
				{
					double err_prob = 1 - (1.0 / grade.num_alignments);
					mapQ = (int)(-10.0 * log(err_prob) / log(10.0));
				}
				else
				{
					mapQ = 255;
				}
				char mapq_buf[64];
				sprintf(mapq_buf, "%d", mapQ);
				sam_toks[t] = mapq_buf;
				break;
			}

			default:
				break;
		}	
		strcat (rebuf, sam_toks[t].c_str());
		if (t != sam_toks.size() - 1)
			strcat(rebuf, "\t");
		else
			strcat(rebuf, "\n");
	}
	return true;
}

bool rewrite_sam_hit(const BowtieHit& bh,
					 const char* bwt_buf,
					 char* rebuf, 
					 char* read_alt_name,
					 const InsertAlignmentGrade& grade,
					 FragmentType insert_side,
					 const BowtieHit* partner)
{
	// Rewrite this hit, filling in the alt name, mate mapping
	// and setting the pair flag
	vector<string> sam_toks;
	tokenize(bwt_buf, "\t", sam_toks);
	char* slash_pos = strrchr(read_alt_name,'/');
	if (slash_pos)
	{
		*slash_pos = 0;
	}
	
	*rebuf = 0;
	
	for (size_t t = 0; t < sam_toks.size(); ++t)
	{
		switch(t)
		{
			case 0: //QNAME
			{
				sam_toks[t] = read_alt_name;
				break;
			}
			case 1: //SAM FLAG
			{
				// 0x0010 (strand of query) is assumed to be set correctly
				// to begin with
				
				int flag = atoi(sam_toks[1].c_str());
				flag |= 0x0001;
				if (insert_side == FRAG_LEFT)
					flag |= 0x0040;
				else if (insert_side == FRAG_RIGHT)
					flag |= 0x0080;
				

				if (grade.happy() && partner)
					flag |= 0x0002;
				
				if (partner)
				{
					if (partner->antisense_align())
						flag |= 0x0020;
				}
				else
				{
					flag |= 0x0008;
				}
				
				char flag_buf[64];
				sprintf(flag_buf, "%d", flag);
				sam_toks[t] = flag_buf;
				break;
			}

			case 4: //MAPQ
			{
				int mapQ;
				if (grade.num_alignments > 1)
				{
					double err_prob = 1 - (1.0 / grade.num_alignments);
					mapQ = (int)(-10.0 * log(err_prob) / log(10.0));
				}
				else
				{
					mapQ = 255;
				}
				char mapq_buf[64];
				sprintf(mapq_buf, "%d", mapQ);
				sam_toks[t] = mapq_buf;
				break;
			}
			case 6: //MRNM
			{
				if (partner)
				{
					//FIXME: this won't be true forever.  Someday, we will report 
					//alignments of pairs not on the same contig.  
					sam_toks[t] = "=";
				}
				else
				{
					sam_toks[t] = "*";
				}
				break;
			}
			case 7: //MPOS
			{
				if (partner)
				{
					char pos_buf[64];
					int partner_pos = partner->left() + 1;  // SAM is 1-indexed
					
					sprintf(pos_buf, "%d", partner_pos);
					sam_toks[t] = pos_buf;
					break;
				}
				else
				{
					sam_toks[t] = "0";
				}
			}
			default:
				break;
		}	
		strcat (rebuf, sam_toks[t].c_str());
		if (t != sam_toks.size() - 1)
			strcat(rebuf, "\t");
		else
			strcat(rebuf, "\n");
	}
	return true;
}

//void print_sam_for_accepted_hits(FILE* fout,
//								 RefSequenceTable& rt,
//								 ReadTable& it,
//								 const string& filename,
//								 FILE* reads_file,
//								 BestFragmentAlignmentTable* best_alignments)
//{
//	typedef pair<FragmentAlignmentGrade, vector<FragmentAlignment*> > FH;
//	
//	HitFactory* hit_factory = NULL;
//
//	rewind(reads_file);
//	
//	hit_factory = new SAMHitFactory(it, rt);
//	
//	FILE* map = fopen(filename.c_str(), "r");
//	
//	if (map == NULL)
//	{
//		fprintf(stderr, "Error: could not open %s\n", filename.c_str());
//		exit(1);
//	}
//	//fprintf(stderr, "Loading hits from %s\n", filename.c_str());
//	//size_t num_hits_before_load = hits.total_hits();
//	
//	static int buf_size = 2048;
//	char bwt_buf[buf_size];
//	uint32_t reads_extracted = 0;
//	
//	char read_name[buf_size];
//	char read_seq[buf_size];
//	char read_alt_name[buf_size];
//	char read_quals[buf_size];
//	
//	ReadID last_id = 0;
//	while (fgets(bwt_buf, 2048, map))
//	{
//		// Chomp the newline
//		char* nl = strrchr(bwt_buf, '\n');
//		if (nl) *nl = 0;
//		
//		if (*bwt_buf == 0)
//			continue;
//		
//		string clean_buf = bwt_buf;
//		// Get a new record from the tab-delimited Bowtie map
//		BowtieHit buf_hit;
//		
//		if (hit_factory->get_hit_from_buf(bwt_buf, buf_hit, true))
//		{
//			if (buf_hit.insert_id() != last_id)
//			{ 
//				get_read_from_stream(buf_hit.insert_id(), 
//									 reads_file,
//									 reads_format,
//									 false,
//									 read_name, 
//									 read_seq,
//									 read_alt_name,
//									 read_quals);
//				last_id = buf_hit.insert_id();
//			}
//			
//			const HitList* hit_list = hits.get_hits(buf_hit.ref_id());
//			if (!hit_list)
//				continue;
//			
//			HitList::const_iterator lb, ub;
//			lb = lower_bound(hit_list->begin(), hit_list->end(), buf_hit, hit_insert_id_lt);
//			ub = upper_bound(hit_list->begin(), hit_list->end(), buf_hit, hit_insert_id_lt);
//			const BowtieHit* processed_hit = NULL;
//			
//			while (lb != ub)
//			{
//				if (*lb == buf_hit && lb->accepted())
//				{
//					processed_hit = &(*lb);
//					break;
//				}
//				++lb;
//			}
//			
//			if (!processed_hit)
//				continue;
//			
//			FH* fh = NULL;
//			if (best_alignments)
//				fh = &((*best_alignments)[buf_hit.insert_id()]);
//			
//			char rebuf[2048];
//			assert (processed_hit->accepted());
//			
//			if (rewrite_sam_hit(*processed_hit, bwt_buf, rebuf, read_alt_name, fh))
//				fprintf(fout, "%s", rebuf);
//
//		}
//		reads_extracted++;
//	}
//    
//}

//void print_sam_for_accepted_hits(FILE* fout,
//								 RefSequenceTable& rt,
//								 ReadTable& it,
//								 const string& filename,
//								 FILE* reads_file)
//{
//	typedef pair<InsertAlignmentGrade, vector<InsertAlignment> > IH;
//
//	HitFactory* hit_factory = NULL;
//
//	rewind(reads_file);
//	
//	hit_factory = new SAMHitFactory(it, rt);
//
//	FILE* map = fopen(filename.c_str(), "r");
//	
//	if (map == NULL)
//	{
//		fprintf(stderr, "Error: could not open %s\n", filename.c_str());
//		exit(1);
//	}
//	//fprintf(stderr, "Loading hits from %s\n", filename.c_str());
//	//size_t num_hits_before_load = hits.total_hits();
//	
//	static int buf_size = 2048;
//	char bwt_buf[buf_size];
//	uint32_t reads_extracted = 0;
//	
//	char read_name[buf_size];
//	char read_seq[buf_size];
//	char read_alt_name[buf_size];
//	char read_quals[buf_size];
//	
//	ReadID last_id = 0;
//	while (fgets(bwt_buf, 2048, map))
//	{
//		// Chomp the newline
//		char* nl = strrchr(bwt_buf, '\n');
//		if (nl) *nl = 0;
//		
//		if (*bwt_buf == 0)
//			continue;
//		
//		string clean_buf = bwt_buf;
//		// Get a new record from the tab-delimited Bowtie map
//		BowtieHit buf_hit;
//		
//		if (hit_factory->get_hit_from_buf(bwt_buf, buf_hit, true))
//		{
//			if (buf_hit.insert_id() != last_id)
//			{ 
//				get_read_from_stream(buf_hit.insert_id(), 
//									 reads_file,
//									 reads_format,
//									 false,
//									 read_name, 
//									 read_seq,
//									 read_alt_name,
//									 read_quals);
//				last_id = buf_hit.insert_id();
//			}
//
//			const HitList* hit_list = hits.get_hits(buf_hit.ref_id());
//			if (!hit_list)
//				continue;
//			
//			HitList::const_iterator lb, ub;
//			lb = lower_bound(hit_list->begin(), hit_list->end(), buf_hit, hit_insert_id_lt);
//			ub = upper_bound(hit_list->begin(), hit_list->end(), buf_hit, hit_insert_id_lt);
//			const BowtieHit* processed_hit = NULL;
//			const BowtieHit* processed_hit_partner = NULL;
//			IH* ih = NULL;
//			
//			while (lb != ub)
//			{
//				if (*lb == buf_hit && lb->accepted())
//				{
//					processed_hit = &(*lb);
//					break;
//				}
//				++lb;
//			}
//			
//			if (!processed_hit)
//				continue;
//			
//			if (best_pairings)
//			{
//				ih = NULL;
//				ih = &((*best_pairings)[buf_hit.insert_id()]);
//				for (size_t p = 0; p < ih->second.size(); ++p)
//				{
//					if (ih->second[p].left_alignment == processed_hit)
//					{
//						processed_hit_partner = ih->second[p].right_alignment;
//						break;
//					}
//					else if (ih->second[p].right_alignment == processed_hit)
//					{
//						processed_hit_partner = ih->second[p].left_alignment;
//						break;
//					}
//				}
//			}
//			
//			char rebuf[2048];
//			assert (processed_hit->accepted());
//
//			if (rewrite_sam_hit(*processed_hit, bwt_buf, rebuf, read_alt_name, ih, processed_hit_partner))
//				fprintf(fout, "%s", rebuf);
//
//		}
//		reads_extracted++;
//	}
//}

//void print_sam_for_accepted_hits(FILE* fout, 
//								 RefSequenceTable& rt,
//								 ReadTable& it,
//								 const string& left_read_maplist,
//								 FILE* left_reads_file,
//								 const string* right_read_maplist,
//								 FILE* right_reads_file,
//								 BestInsertAlignmentTable* best_pairings)
//{
//    vector<string> left_filenames;
//    tokenize(left_read_maplist, ",", left_filenames);
//    print_sam_for_accepted_hits(fout, rt, it, left_filenames, left_reads_file, best_pairings);
//    if (right_read_maplist)
//    {
//        vector<string> right_filenames;
//        tokenize(*right_read_maplist, ",", right_filenames);
//        print_sam_for_accepted_hits(fout, rt, it, right_filenames, right_reads_file, best_pairings);
//    }
//}

//void print_sam_for_accepted_hits(FILE* fout, 
//								 RefSequenceTable& rt,
//								 ReadTable& it,
//								 const string& left_read_maplist,
//								 FILE* left_reads_file,
//								 BestFragmentAlignmentTable* best_alignments)
//{
//    vector<string> left_filenames;
//    tokenize(left_read_maplist, ",", left_filenames);
//    print_sam_for_accepted_hits(fout, rt, it, left_filenames, left_reads_file, best_alignments);
//}

void print_sam_for_hits(const HitsForRead& hits,
						const FragmentAlignmentGrade& grade,
						FragmentType frag_type,
						FILE* reads_file,
						FILE* fout)
{
	static const int buf_size = 2048;
	char read_name[buf_size];
	char read_seq[buf_size];
	char read_alt_name[buf_size];
	char read_quals[buf_size];
	
	char rebuf[buf_size];
	
	bool got_read = get_read_from_stream(hits.insert_id, 
										 reads_file,
										 reads_format,
										 false,
										 read_name, 
										 read_seq,
										 read_alt_name,
										 read_quals);
	
	assert (got_read);
	
	for (size_t i = 0; i < hits.hits.size(); ++i)
	{
		const BowtieHit& bh = hits.hits[i];
		if (rewrite_sam_hit(bh, bh.hitfile_rec().c_str(), rebuf, read_alt_name, grade, frag_type))
			fprintf(fout, "%s", rebuf);
	}
}

void print_sam_for_hits(const HitsForRead& left_hits,
						const HitsForRead& right_hits,
						const InsertAlignmentGrade& grade,
						FILE* left_reads_file,
						FILE* right_reads_file,
						FILE* fout)
{
	assert (left_hits.insert_id == right_hits.insert_id);
	
	static const int buf_size = 2048;
	char left_read_name[buf_size];
	char left_read_seq[buf_size];
	char left_read_alt_name[buf_size];
	char left_read_quals[buf_size];
	char left_rebuf[buf_size];
	
	char right_read_name[buf_size];
	char right_read_seq[buf_size];
	char right_read_alt_name[buf_size];
	char right_read_quals[buf_size];
	char right_rebuf[buf_size];
	
	bool got_left_read = get_read_from_stream(left_hits.insert_id, 
											  left_reads_file,
											  reads_format,
											  false,
											  left_read_name, 
											  left_read_seq,
											  left_read_alt_name,
											  left_read_quals);
	
	bool got_right_read = get_read_from_stream(right_hits.insert_id, 
											   right_reads_file,
											   reads_format,
											   false,
											   right_read_name, 
											   right_read_seq,
											   right_read_alt_name,
											   right_read_quals);
	
	assert (left_hits.hits.size() == right_hits.hits.size() ||
			(left_hits.hits.empty() || right_hits.hits.empty()));
	
	if (left_hits.hits.size() == right_hits.hits.size())
	{
		assert (got_left_read && got_right_read);
		for (size_t i = 0; i < right_hits.hits.size(); ++i)
		{
			const BowtieHit& right_bh = right_hits.hits[i];
			if (rewrite_sam_hit(right_bh, 
								right_bh.hitfile_rec().c_str(), 
								right_rebuf, 
								right_read_alt_name, 
								grade, 
								FRAG_RIGHT, 
								&left_hits.hits[i]))
			{
				fprintf(fout, "%s", right_rebuf);
			}
			
			const BowtieHit& left_bh = left_hits.hits[i];
			if (rewrite_sam_hit(left_bh, 
								left_bh.hitfile_rec().c_str(), 
								left_rebuf, 
								left_read_alt_name, 
								grade, 
								FRAG_LEFT, 
								&right_hits.hits[i]))
			{
				fprintf(fout, "%s", left_rebuf);
			}
		}
	}
	else if (left_hits.hits.empty())
	{
		for (size_t i = 0; i < right_hits.hits.size(); ++i)
		{
			const BowtieHit& bh = right_hits.hits[i];
			if (rewrite_sam_hit(bh, 
								bh.hitfile_rec().c_str(), 
								right_rebuf, 
								right_read_alt_name, 
								grade, 
								FRAG_RIGHT, 
								NULL))
				fprintf(fout, "%s", right_rebuf);
		}
	}
	else if (right_hits.hits.empty())
	{
		for (size_t i = 0; i < left_hits.hits.size(); ++i)
		{
			const BowtieHit& bh = left_hits.hits[i];
			if (rewrite_sam_hit(bh, 
								bh.hitfile_rec().c_str(), 
								left_rebuf, 
								left_read_alt_name, 
								grade, 
								FRAG_LEFT, 
								NULL))
				fprintf(fout, "%s", left_rebuf);
		}
	}
	else
	{
		assert (false);
	}
}

// Extracts junctions from all the SAM hits (based on REF_SKIPs) in the hit file
// resets the stream when finished.
void get_junctions_from_hits(HitStream& hit_stream, 
							 ReadTable& it, 
							 JunctionSet& junctions)
{
	HitsForRead curr_hit_group;
	hit_stream.next_read_hits(curr_hit_group);

	uint32_t curr_obs_order = it.observation_order(curr_hit_group.insert_id);
	
	while(curr_obs_order != 0xFFFFFFFF)
	{
		for (size_t i = 0; i < curr_hit_group.hits.size(); ++i)
		{
			BowtieHit& bh = curr_hit_group.hits[i];
			if (!bh.contiguous())
			{
				junctions_from_alignment(bh, junctions);
			}
			hit_stream.next_read_hits(curr_hit_group);
			curr_obs_order = it.observation_order(curr_hit_group.insert_id);
		}
	}
	
	hit_stream.reset();
}

void update_junctions(const HitsForRead& hits,
						   JunctionSet& junctions)
{
	for (size_t i = 0; i < hits.hits.size(); ++i)
	{
		const BowtieHit& bh = hits.hits[i];
		junctions_from_alignment(bh, junctions);
	}
}

// Extracts junctions from all the SAM hits (based on REF_SKIPs) in the hit file
// resets the stream when finished.
void exclude_hits_on_filtered_junctions(const JunctionSet& junctions,
										HitsForRead& hits)
{
	HitsForRead remaining;
	remaining.insert_id = hits.insert_id;

	for (size_t i = 0; i < hits.hits.size(); ++i)
	{
		BowtieHit& bh = hits.hits[i];
		bool filter_hit = false;
		if (!bh.contiguous())
		{
			JunctionSet bh_juncs;
			junctions_from_alignment(bh, bh_juncs);
			for (JunctionSet::iterator itr = bh_juncs.begin();
				 itr != bh_juncs.end();
				 itr++)
			{
				const Junction& j = itr->first;
				if (junctions.find(j) == junctions.end())
				{
					filter_hit = true;
					break;
				}
			}
		}
		if (!filter_hit)
		{
			remaining.hits.push_back(bh);
		}
		else
		{
			//fprintf(stderr, "Filtering hit\n");
		}
	}
	hits = remaining;
}


void driver(FILE* left_map,
			FILE* left_reads,
            FILE* right_map,
			FILE* right_reads,
            FILE* junctions_out,
			FILE* accepted_hits_out)
{	
    ReadTable it;
    RefSequenceTable rt(true);
	
    SAMHitFactory hit_factory(it,rt);
	
	HitStream left_hs(left_map, &hit_factory, false, true, true);
	HitStream right_hs(right_map, &hit_factory, false, true, true);
	
    JunctionSet junctions;
		
	get_junctions_from_hits(left_hs, it, junctions);
	get_junctions_from_hits(right_hs, it, junctions);
	
	HitsForRead curr_left_hit_group;
	HitsForRead curr_right_hit_group;
	
	left_hs.next_read_hits(curr_left_hit_group);
	right_hs.next_read_hits(curr_right_hit_group);

	uint32_t curr_left_obs_order = it.observation_order(curr_left_hit_group.insert_id);
	uint32_t curr_right_obs_order = it.observation_order(curr_right_hit_group.insert_id);
	
	// Read hits, extract junctions, and toss the ones that arent strongly enough supported.
	
	filter_junctions(junctions);
	
	JunctionSet final_junctions; // the junctions formed from best hits
	
	fprintf (stderr, "Reporting final accepted alignments...");
	
	// While we still have unreported hits...
	while(curr_left_obs_order != 0xFFFFFFFF || 
		  curr_right_obs_order != 0xFFFFFFFF)
	{		
		// Chew up left singletons
		while (curr_left_obs_order < curr_right_obs_order &&
			   curr_left_obs_order != 0xFFFFFFFF)
		{
			HitsForRead best_hits;
			best_hits.insert_id = curr_left_obs_order;
			FragmentAlignmentGrade grade;
			
			exclude_hits_on_filtered_junctions(junctions, curr_left_hit_group);
			
			// Process hits for left singleton, select best alignments
			fragment_best_alignments(curr_left_hit_group, junctions, grade, best_hits);
			
			update_junctions(best_hits, final_junctions);
			
			print_sam_for_hits(best_hits, 
							   grade,
							   right_map ? FRAG_LEFT : FRAG_UNPAIRED,
							   left_reads, 
							   accepted_hits_out);
			
			// TODO: update junctionstats to reflect best hits
			
			// Get next hit group
			left_hs.next_read_hits(curr_left_hit_group);
			curr_left_obs_order = it.observation_order(curr_left_hit_group.insert_id);
		}
		
		// Chew up right singletons
		while (curr_left_obs_order > curr_right_obs_order &&
			   curr_right_obs_order != 0xFFFFFFFF)
		{
			HitsForRead best_hits;
			best_hits.insert_id = curr_right_obs_order;
			FragmentAlignmentGrade grade;
			
			exclude_hits_on_filtered_junctions(junctions, curr_right_hit_group);
			
			// Process hit for right singleton, select best alignments
			fragment_best_alignments(curr_right_hit_group, junctions, grade, best_hits);
			
			update_junctions(best_hits, final_junctions);
			
			print_sam_for_hits(best_hits, 
							   grade, 
							   FRAG_RIGHT,
							   right_reads, 
							   accepted_hits_out);
			
			// TODO: update junctionstats to reflect best hits
			
			// Get next hit group
			right_hs.next_read_hits(curr_right_hit_group);
			curr_right_obs_order = it.observation_order(curr_right_hit_group.insert_id);
		}
		
		// Since we have both left hits and right hits for this insert,
		// Find the best pairing and print both
		while (curr_left_obs_order == curr_right_obs_order &&
			   curr_left_obs_order != 0xFFFFFFFF && curr_right_obs_order != 0xFFFFFFFF)
		{			 
			HitsForRead left_best_hits;
			HitsForRead right_best_hits;
			
			left_best_hits.insert_id = curr_left_obs_order;
			right_best_hits.insert_id = curr_right_obs_order;
			InsertAlignmentGrade grade;
			
			exclude_hits_on_filtered_junctions(junctions, curr_left_hit_group);
			exclude_hits_on_filtered_junctions(junctions, curr_right_hit_group);
			
			if (curr_left_hit_group.hits.empty())
			{
				FragmentAlignmentGrade grade;
				fragment_best_alignments(curr_right_hit_group, junctions, grade, right_best_hits);
				
				update_junctions(right_best_hits, final_junctions);
				
				print_sam_for_hits(right_best_hits, 
								   grade, 
								   FRAG_RIGHT,
								   right_reads, 
								   accepted_hits_out);	
			}
			else if (curr_right_hit_group.hits.empty())
			{
				FragmentAlignmentGrade grade;
				// Process hits for left singleton, select best alignments
				fragment_best_alignments(curr_left_hit_group, junctions, grade, left_best_hits);
				
				update_junctions(left_best_hits, final_junctions);
				
				print_sam_for_hits(left_best_hits, 
								   grade,
								   FRAG_LEFT,
								   left_reads, 
								   accepted_hits_out);
			}
			else
			{				
				insert_best_alignments(curr_left_hit_group, 
									   curr_right_hit_group, 
									   junctions,
									   grade,
									   left_best_hits,
									   right_best_hits);
				
				update_junctions(left_best_hits, final_junctions);
				update_junctions(right_best_hits, final_junctions);
				
				print_sam_for_hits(left_best_hits,
								   right_best_hits,
								   grade,
								   left_reads, 
								   right_reads, 
								   accepted_hits_out);
			}
			
			// TODO: update junctionstats to reflect best hits
			
			left_hs.next_read_hits(curr_left_hit_group);
			curr_left_obs_order = it.observation_order(curr_left_hit_group.insert_id);
			
			right_hs.next_read_hits(curr_right_hit_group);
			curr_right_obs_order = it.observation_order(curr_right_hit_group.insert_id);
		}
	}
	
	for (JunctionSet::iterator i = junctions.begin(); i != junctions.end();)
	{
		
		if (i->second.supporting_hits > 0)
		{
			++i;
		}
		else
		{
			junctions.erase(i++);
		}
	}
	
	fprintf (stderr, "Printing junction BED track...");
	print_junctions(junctions_out, final_junctions, rt);
	fprintf (stderr, "done\n");
	
    fprintf(stderr, "Found %lu junctions from happy spliced reads\n", junctions.size());
}

void print_usage()
{
    fprintf(stderr, "Usage:   tophat_reports <junctions.bed> <accepted_hits.sam> <left_map1,...,left_mapN> <left_reads.fq>  [right_map1,...,right_mapN] [right_reads.fq]\n");
	
	//    fprintf(stderr, "Usage:   tophat_reports <coverage.wig> <junctions.bed> <accepted_hits.sam> <map1.bwtout> [splice_map1.sbwtout]\n");
}

int main(int argc, char** argv)
{
	fprintf(stderr, "tophat_reports v%s\n", PACKAGE_VERSION); 
	fprintf(stderr, "---------------------------------------\n");
	
	reads_format = FASTQ;
	
    int parse_ret = parse_options(argc, argv, print_usage);
    if (parse_ret)
        return parse_ret;

    if(optind >= argc)
    {
        print_usage();
        return 1;
    }

    string junctions_file_name = argv[optind++];

    if(optind >= argc)
    {
        print_usage();
        return 1;
    }
	
	string accepted_hits_file_name = argv[optind++];
	
    if(optind >= argc)
    {
        print_usage();
        return 1;
    }

    string left_map_filename = argv[optind++];
	
	if(optind >= argc)
    {
        print_usage();
        return 1;
    }
	
	FILE* left_map = fopen(left_map_filename.c_str(), "r");
	if (!left_map)
	{
		fprintf(stderr, "Error: cannot open map file %s for reading\n",
				left_map_filename.c_str());
		exit(1);
	}
	
    string left_reads_filename = argv[optind++];
	
    string* right_map_filename = NULL;
	FILE* right_map = NULL;
	string* right_reads_filename = NULL;
	FILE* right_reads_file = NULL;
	
    if (optind < argc)
	{
        right_map_filename = new string(argv[optind++]);
		
		if(optind >= argc)
		{
			print_usage();
			return 1;
		}
		
		right_map = fopen(right_map_filename->c_str(), "r");
		if (!right_map)
		{
			fprintf(stderr, "Error: cannot open map file %s for reading\n",
					right_map_filename->c_str());
			exit(1);
		}
		
		right_reads_filename = new string(argv[optind++]);
		right_reads_file = fopen(right_reads_filename->c_str(), "r");
		if (!right_reads_file)
		{
			fprintf(stderr, "Error: cannot open reads file %s for reading\n",
					right_reads_filename->c_str());
			exit(1);
		}
	}

    FILE* junctions_file = fopen((output_dir + "/" + junctions_file_name).c_str(), "w");
    if (junctions_file == NULL)
    {
        fprintf(stderr, "Error: cannot open BED file %s for writing\n",
                junctions_file_name.c_str());
        exit(1);
    }

    FILE* accepted_hits_file = fopen((output_dir + "/" + accepted_hits_file_name).c_str(), "w");
    if (accepted_hits_file == NULL)
    {
        fprintf(stderr, "Error: cannot open SAM file %s for writing\n",
                accepted_hits_file_name.c_str());
        exit(1);
    }
	
	FILE* left_reads_file = fopen(left_reads_filename.c_str(), "r");
    if (!left_reads_file)
    {
        fprintf(stderr, "Error: cannot open reads file %s for reading\n",
                left_reads_filename.c_str());
        exit(1);
    }

    driver(left_map,
		   left_reads_file,
           right_map,
		   right_reads_file,
           junctions_file,
		   accepted_hits_file);

    return 0;
}


